#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
import shutil
import os
import numpy as np
from astropy.io import fits
from astropy import wcs
from astropy.table import Table, vstack
from shapely.geometry import Polygon

from muphoten.utils import (rm_p, load_config, getpath)
from muphoten.astrometry import scamp


def get_crpix(proj_crpix1, proj_crpix2, Xcell, Ycell, x, y):
    """
    Compute CRPIX1 and CRPIX2 for cell based on the CRPIX values
    of the projcell.
    """
    x_center, y_center = 5, 5
    cprix1 = proj_crpix1 + (x_center - x) * (Xcell - 480)
    crpix2 = proj_crpix2 + (y_center - y) * (Ycell - 480)

    return cprix1, crpix2


def get_RADEC_coord(proj_crpix1, proj_crpix2, Xcell, Ycell, x, y, RA, Dec):

    pixscale = 0.25 / 3600
    crpix1, crpix2 = get_crpix(proj_crpix1, proj_crpix2, Xcell, Ycell, x, y)

    # Create a new WCS object.  The number of axes must be set
    # from the start
    w = wcs.WCS(naxis=2)

    # Set up projection
    # Vector properties may be set with Python lists, or Numpy arrays
    w.wcs.crpix = [float(crpix1), float(crpix2)]
    w.wcs.cdelt = np.array([-pixscale, pixscale])
    w.wcs.crval = [RA, Dec]
    w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
    # w.wcs.set_pv([(2, 1, 45.0)])

    # Three pixel coordinates of interest.
    # The pixel coordinates are pairs of [X, Y].
    # The "origin" argument indicates whether the input coordinates
    # are 0-based (as in Numpy arrays) or
    # 1-based (as in the FITS convention, for example coordinates
    # coming from DS9).
    pixcrd = np.array(
        [[0, 0], [Xcell, 0], [Xcell, Ycell], [0, Ycell]], dtype=np.float64
    )

    # Convert pixel coordinates to world coordinates.
    # The second argument is "origin" -- in this case we're declaring we
    # have 0-based (Numpy-like) coordinates.
    world = w.wcs_pix2world(pixcrd, 1)
    world = np.array(world).T

    return world, w


def ps1_cell_coord(im_corner_coords, projcell_id, Xcell, Ycell,
                   projcell_ra_center, projcell_dec_center,
                   proj_crpix1, proj_crpix2):
    """
    Template for computing the 10x10 cells composing a PS1 projcell
    """
    #  cell size, 0yx
    ny = 10  #  bottom to top of projell (ascending dec)
    nx = 10  # left to right of projcell (ascending ra)

    # Need to append a 0 when < 1000 otherwise download crashes
    if projcell_id < 1000:
        projcell_id = "0" + str(projcell_id)
    else:
        projcell_id = str(projcell_id)

    id_list = []
    RA_min = []
    RA_max = []
    dec_min = []
    dec_max = []
    projcell_id_list = []
    for y in range(ny):
        for x in range(nx):
            #  Estimate corner coordinates for each cell
            corner_coords, w = get_RADEC_coord(
                proj_crpix1,
                proj_crpix2,
                Xcell,
                Ycell,
                x,
                y,
                projcell_ra_center,
                projcell_dec_center,
            )
            #  Check whether one cell is contained in the input image
            pix_im_coord = np.array(
                [im_corner_coords[0], im_corner_coords[1]]).T
            pix_cell_coord = np.array([corner_coords[0], corner_coords[1]]).T

            #  Check for cells having RA around below and above 0 degrees
            cell_RAs = pix_cell_coord[:, 0]
            cell_RA_max = np.max(cell_RAs)
            flag = False
            #  if one corner has distance more than 300 degrees
            #  to highest RA, add 360 to it
            #  and to the image RA corners
            #  So that intersect with shapely still make sense
            # Might need to change for very large field of view.
            for i, cell_RA in enumerate(cell_RAs):
                if cell_RA_max - cell_RA > 300:
                    flag = True
                    pix_cell_coord[i, 0] += 360
            if flag:
                pix_im_coord[:, 0] += 360

            im_poly = Polygon([tuple(co) for co in pix_im_coord])
            cell_poly = Polygon([tuple(co) for co in pix_cell_coord])
            # print (polygon_im.contains(cell_corner_coords, w))
            if im_poly.intersects(cell_poly):
                projcell_id_list.append(projcell_id)
                id_list.append("0%d%d" % (y, x))
                RA_min.append(np.min(corner_coords[0], axis=0))
                RA_max.append(np.max(corner_coords[0], axis=0))
                dec_min.append(np.min(corner_coords[1], axis=0))
                dec_max.append(np.max(corner_coords[1], axis=0))

    overlap_cells = Table(
        [projcell_id_list, id_list, RA_min, RA_max, dec_min, dec_max],
        names=(
            "projcell_id",
            "cell_id",
            "RA_min",
            "RA_max",
            "dec_min",
            "dec_max"),
    )
    return overlap_cells

def zone_PS1(ps1grid, cellID):
    """Returns the ZONE in which a cellID is."""

    if cellID < ps1grid["PROJCELL"][0] or \
            cellID > ps1grid["PROJCELL"][-1]:
        errmsg = 'Cell ID %d is outside the expected range: [%d-%d]' % \
                (cellID, ps1grid["PROJCELL"][0], ps1grid["PROJCELL"][-1])
        raise BaseException(errmsg)

    # If last zone, it contains only one cell so easy
    if cellID == 2643:
        zone = 45

    for i in range(len(ps1grid) - 1):
        if ps1grid["PROJCELL"][i] <= cellID and \
                ps1grid["PROJCELL"][i+1] > cellID:
            zone = ps1grid["ZONE"][i]

    return zone

def ps1_grid(im_corner_coords):
    """
    Return the ps1 cell IDs for a given image dimension
    Skycell images have names like skycell.nnnn.0yx where nnnn
    is the projection cell number (which ranges from 635 to 2643)
    and 0yx gives the skycell location in the image, with y and x
    ranging from 0 to 9 indicating the respective y and x section
    of the projection cell. The 000 skycell is in the bottom left
    corner of the projection cell, 010 is just above it, and 099
    is in the upper right corner.
    RA = n *360 deg / M
    """
    path_muphoten = getpath()

    # RA, dec min and max of input image
    ra_min = np.min(im_corner_coords[0], axis=0)
    ra_max = np.max(im_corner_coords[0], axis=0)
    dec_min = np.min(im_corner_coords[1], axis=0)
    dec_max = np.max(im_corner_coords[1], axis=0)

    ps1grid = Table.read(path_muphoten + "/ps1_survey/ps1grid.fits", hdu=1)
    #  Get the min and max declination zones
    mask = ((ps1grid["DEC_MIN"] < dec_min) & (ps1grid["DEC_MAX"] > dec_min)) | (
        (ps1grid["DEC_MIN"] < dec_max) & (ps1grid["DEC_MAX"] > dec_max)
    )
    # Get all declinations zones
    all_zones_id = np.arange(
        ps1grid[mask]["ZONE"][0], ps1grid[mask]["ZONE"][-1] + 1)
    all_cells = []

    # Loop over the different zones
    for zone in all_zones_id:
        mask = ps1grid["ZONE"] == zone
        idx_bkp = -1
        projcell_idx_list = []
        for ra in [ra_min, ra_max]:
            #  Get the cells covering the input ra
            closet_projcell_idx = (
                float(ps1grid[mask]["PROJCELL"])
                + ra * float(ps1grid[mask]["NBAND"]) / 360
            )
            projcell_idx = int(np.floor(closet_projcell_idx))
            if projcell_idx != idx_bkp:
                projcell_idx_list.append(projcell_idx)
            idx_bkp = projcell_idx
        #  Check the 0 - 360 transition
        #  Assume that an image can not have a FoV larger than 60 degrees
        #  Need to do it dependng on the image in the future
        if ra_max - ra_min < 300:
            total_proj_cell_idx = np.arange(
                projcell_idx_list[0], projcell_idx_list[-1] + 1
            )
        else:
            #  Image overlapping the 0 - 360 region
            total_proj_cell_idx = np.arange(
                ps1grid[mask]["PROJCELL"], projcell_idx_list[0] + 1
            )
            list_proj_cell_end = np.arange(
                projcell_idx_list[-1], ps1grid[ps1grid["ZONE"]
                                               == zone + 1]["PROJCELL"]
            )
            total_proj_cell_idx = np.append(
                total_proj_cell_idx, list_proj_cell_end)

        for cell_id in total_proj_cell_idx:
            mask = ps1grid["ZONE"] == zone_PS1(ps1grid, cell_id)
            diff_projcell_idx = cell_id - float(ps1grid[mask]["PROJCELL"])
            ra_center_projcell = diff_projcell_idx * \
                360 / float(ps1grid[mask]["NBAND"])
            cell_query = ps1_cell_coord(
                im_corner_coords,
                cell_id,
                ps1grid[mask]["XCELL"],
                ps1grid[mask]["YCELL"],
                ra_center_projcell,
                ps1grid[mask]["DEC"],
                ps1grid[mask]["CRPIX1"],
                ps1grid[mask]["CRPIX2"],
            )

            if len(cell_query) > 0:
                all_cells.append(cell_query)

    if len(all_cells) == 1:
        all_cells = all_cells[0]
    else:
        all_cells = vstack([tab for tab in all_cells])
    return all_cells


def download_ps1_cells(cell_table, band, config, ps1Dir,
                       ps1RescaledDir, verbose="QUIET"):
    """Download the required cell from PS1 DR1"""

    file_list = []
    #  extension to get auxiliary images
    # See https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images
    # auxiliaryFiles = ['.mask', '.wt', '.num', '.exp', '.expwt', '']
    auxiliaryFiles = ["", ".mask"]

    BaseURL = "http://ps1images.stsci.edu/"
    # cell_table = [cell_table[7]]
    for cell in cell_table:
        cell_url_path = "rings.v3.skycell/%s/%s/" % (
            cell["projcell_id"],
            cell["cell_id"],
        )
        for aux in auxiliaryFiles:
            cell_file = "rings.v3.skycell.%s.%s.stk.%s.unconv%s.fits" % (
                cell["projcell_id"],
                cell["cell_id"],
                band,
                aux,
            )
            Link = BaseURL + cell_url_path + cell_file
            local_cell_file = cell_file.replace(
                ".", "_").replace("_fits", ".fits")
            FileNameFitsPath = ps1Dir + local_cell_file
            if os.path.isfile(FileNameFitsPath):
                print("File %s already downloaded" % FileNameFitsPath)
            else:
                # wget_command = "wget %s -O %s"%(Link,FileNameFitsPath)
                wget_command = "curl -m 7200 -L -o %s %s" % (
                    FileNameFitsPath, Link)
                os.system(wget_command)
                if os.path.isfile(FileNameFitsPath):
                    # do not really understand what this is doing
                    funpack_command = "fpack %s; rm %s; funpack %s.fz" % (
                        FileNameFitsPath,
                        FileNameFitsPath,
                        FileNameFitsPath,
                    )
                    os.system(funpack_command)

                    rm_command = "rm %s.fz" % (FileNameFitsPath)
                    os.system(rm_command)
                else:
                    print(
                        "File %s was not downloaded or found on the server." %
                        Link)

            new_filename = local_cell_file.replace(
                    '.fits', '_%s.fits' % config['telescope']
            )
            if aux == "":
                # Check if targeted file was downloaded to continue
                if os.path.isfile(FileNameFitsPath):
                    if os.path.isfile(ps1RescaledDir + new_filename):
                       pass
                    else:
                        #  Rescale to physical flux
                        linear_rescale_ps1(
                            local_cell_file, ps1Dir, ps1RescaledDir, band
                        )
                        resample_ps1(local_cell_file,
                                     ps1RescaledDir,
                                     config)
                        #  Perform astrometric calibration on each cell
                        config_ps1 = load_config("PS1", "default")
                        scamp(
                            ps1RescaledDir + new_filename,
                            config_ps1,
                            accuracy=0.1,
                            itermax=3,
                            band=band,
                            verbose="QUIET",
                        )
            else:
                if os.path.isfile(ps1RescaledDir + new_filename):
                    pass
                else:
                    resample_ps1(local_cell_file,
                                 ps1RescaledDir,
                                 config)
                    # No need to run scamp, pixels are already aligned with
                    # PS1 image which had correct wcs. 
            # new_filename contains the telescope alias.
            file_list.append(ps1RescaledDir + new_filename)
    return file_list


def prepare_PS1_sub(ps1_cell_table, band, inputimage,
                    config, verbose="QUIET", method="individual"):
    """Prepare the download and formatting of PS1 images to be used for image
    substraction"""

    path_muphoten = getpath()
    path, filenameInput = os.path.split(inputimage)
    if path:
        folder = path + "/"
    else:
        folder = ""

    # Get pixelscale from input science image
    # And add it to config dictionary
    header = fits.getheader(inputimage)
    if "CDELT1" in header:
        pixScale = [abs(float(header["CDELT1"])) * 3600,
                    abs(float(header["CDELT2"])) * 3600]
    else:
        pixScale = [np.sqrt(float(header["CD1_1"])**2 +
                            float(header["CD2_1"]**2)) * 3600,
                    np.sqrt(float(header["CD1_2"])**2 +
                            float(header["CD2_2"]**2)) * 3600]

    config['pixScale'] = pixScale

    ps1Dir = path_muphoten + "/ps1Dir/"
    if not os.path.isdir(ps1Dir):
        os.makedirs(ps1Dir)

    ps1RescaledDir = path_muphoten + "/ps1RescaledDir/"
    if not os.path.isdir(ps1RescaledDir):
        os.makedirs(ps1RescaledDir)

    #  Download PS1 files if not present, and reformat them
    ps1files = download_ps1_cells(
        ps1_cell_table, band, config, ps1Dir, ps1RescaledDir, verbose=verbose
    )
    subfiles = []
    if method == "mosaic":
        #  Create mosaic file if it does not exist
        mosaicfile = folder + os.path.splitext(filenameInput)[0] + "_ps1_mosaic.fits"
        if os.path.isfile(mosaicfile):
            print(
                "PS1 mosaic image already exists in this location: %s."
                "If you want to recompute it, delete it."
                % mosaicfile
            )
        else:
            fileref_names = create_ps1_mosaic(
                ps1files, inputimage, folder, config, band, verbose=verbose
            )
            subfiles.append([inputimage, fileref_names[0], fileref_names[1]])
    elif method == "individual":
        for i in range(0, len(ps1files), 2):
            ref = ps1files[i]
            mask = ps1files[i + 1]
            subfiles.append([inputimage, ref, mask])
    return subfiles


def linear_rescale_ps1(filename, inputDir, outputDir,
                       band, normalise=True, method="headers"):
    """rescale the PS1 DR1 fits file"""
    # Transform into linear flux scale
    hdulist = fits.open(inputDir + filename)

    boffset = hdulist[0].header["BOFFSET"]
    bsoften = hdulist[0].header["BSOFTEN"]
    a = 2.5 / np.log(10)
    hdulist[0].data = boffset + 2 * bsoften * np.sinh(hdulist[0].data / a)

    #  Normalise to 1s exposure time
    if normalise:
        #  'exposure_map', exptime, 'headers'
        if method == "exptime":
            print("Use exptime in header to rescale to an exposure of 1s.")
            exptime = float(hdulist[0].header["EXPTIME"])
            hdulist[0].data /= exptime

            try:
                # hdulist[0].header['SATURATE'] /= hdulist[0].header['EXPTIME']
                hdulist[0].header["SATURATE"] /= exptime
            except BaseException:
                pass

        if method == "headers":
            print("Use header information to rescale to an exposure of 1s.")
            #  Check for SCL_* keywords. It corresponds to the scaling factor
            #  applied to each individual exposure. If 0 it is not taken into
            #  in the stacked image.
            #  So exposure time is weighted by this factor:
            #  0 if SCL_*=0
            #  1 if SCL_*>0
            hdr = hdulist[0].header
            SCLlist = hdr["SCL_*"]
            scale_flux = []
            for SCL in SCLlist:
                if float(hdr[SCL]) > 0:
                    scale_flux.append(1)
                else:
                    scale_flux.append(0)
            explist = hdr["EXP_*"]
            exptime_tot = 0
            for i, exp in enumerate(explist):
                exptime_tot += float(scale_flux[i]) * float(hdr[exp])
            # hdulist[0].data /= hdulist[0].header['EXPTIME']
            hdulist[0].data /= exptime_tot

            try:
                # hdulist[0].header['SATURATE']/= hdulist[0].header['EXPTIME']
                hdulist[0].header["SATURATE"] /= exptime_tot
            except BaseException:
                pass

        elif method == "exposure_map":
            print("Use exposure map to rescale to an exposure of 1s.")
            #  Normalise by exact exposure time in each pixels
            # hdulist_exp=fits.open(
            #   inputDir+filename.split('.')[0]+'_exp.fits')
            # hdulist[0].data = hdulist[0].data / hdulist_exp[0].data
            hdulist_expwt = fits.open(
                inputDir + os.path.splitext(filename)[0] + "_expwt.fits")
            hdulist[0].data = hdulist[0].data / hdulist_expwt[0].data
        hdulist[0].header["EXPTIME"] = 1
    """
    # Add some keywords for performing photometric calibration with scamp
    hdulist[0].header['FILTER'] = band
    hdulist[0].header['PHOT_C'] = 25
    hdulist[0].header['PHOT_K'] = 0
    hdulist[0].header['PHOTFLAG'] = 'T'
    """
    hdulist.writeto(outputDir + filename, overwrite=True)
    hdulist.close()

    # replace pixels == 0 with NaNs. Mostly the border, saturated pixels
    hdulist = fits.open(outputDir + filename)
    hdulist[0].data[hdulist[0].data == 0] = np.nan
    hdulist.writeto(outputDir + filename, overwrite=True)

    #  Create a mask to propagate the nan pixels
    hdulist = fits.open(outputDir + filename)
    hdulist[0].data[np.isfinite(hdulist[0].data)] = 0
    hdulist[0].data[np.isnan(hdulist[0].data)] = 1
    hdulist.writeto(
        outputDir +
        os.path.splitext(filename)[0] +
        "_mask.fits",
        overwrite=True)

    return True

def resample_ps1(filename, inputDir, config, useweight=False,
                 verbose="NORMAL"):
    """Resample PS1 image to the telescope resolution"""
    pixScale = config['pixScale'][0]

    basename = os.path.splitext(filename)[0]
    if useweight:
        subprocess.call(
            [
             "swarp",
             inputDir+filename,
             "-IMAGEOUT_NAME", inputDir + basename + \
                         '_%s' % config['telescope'] + ".fits",
             "-WEIGHTOUT_NAME", inputDir + epoch + \
                         '_%s' % config['telescope'] + ".weight.fits",
             "-VERBOSE_TYPE", verbose,
            ]
        )
    else:
        # Use bilinear to avoid artefact, but worst for
        # noise so would need to check in more details.
        subprocess.call(
           [
            "swarp",
            inputDir+filename,
            "-IMAGEOUT_NAME", inputDir + basename + \
                       '_%s' % config['telescope'] + ".fits",
            # '-GAIN_DEFAULT', str(gain),
            "-FSCALE_KEYWORD", "NONE",
            "-FSCALE_DEFAULT", "1, 1",
            "-SUBTRACT_BACK", "N",
            "-COMBINE", "N",
            "-BACK_SIZE", "128",
            "-BACK_FILTERSIZE", "3",
            "-RESAMPLE", "Y",
            "-RESAMPLE_DIR", inputDir,
            "-RESAMPLE_SUFFIX", '_%s.fits' % config['telescope'],
            "-PIXELSCALE_TYPE", "MANUAL",
            "-PIXEL_SCALE", str(pixScale),
            # '-CENTER', '%s, %s' % (header['CRVAL1'],
            #                         header['CRVAL2']),
            # '-RESAMPLING_TYPE', 'LANCZOS3',
            "-RESAMPLING_TYPE", "BILINEAR",
            # '-RESAMPLING_TYPE', 'NEAREST',
            "-OVERSAMPLING", "0",
            "-COMBINE_TYPE", "MEDIAN",
            "-VERBOSE_TYPE", verbose,
            "-COPY_KEYWORDS", "FILTER, EXPTIME, SATURATE",
           ]
        )
    rm_p('swarp.xml')
    rm_p("coadd.weight.fits")
    rm_p(inputDir+filename)

    
    if 'mask' in basename:
        # Add the new pixels created on the edge after resampling to
        # the mask
        hdulist1 = fits.open(inputDir + basename + \
                '_%s' % config['telescope'] + ".weight.fits")
        zero_pix = hdulist1[0].data == 0
        hdulist1.close()

        outName = inputDir + basename + \
                '_%s' % config['telescope'] + ".fits"
        hdulist2 = fits.open(outName)
        # Put high value
        hdulist2[0].data[zero_pix] = 1e8
        hdulist2.writeto(outName, overwrite=True)

    return True

def create_ps1_mosaic(file_list, inputimage, outputDir, config,
                      band, useweight=False, verbose="NORMAL"):
    """Create a single mosaic of PS1 image using swarp"""
    _, filenameInput = os.path.split(inputimage)

    #  Create list of mask fits
    ima_list = [ima for ima in file_list if "_mask" not in ima]
    mask_list = [ima for ima in file_list if "_mask" in ima]
    np.savetxt("mosaic.list", ima_list, fmt="%s")
    np.savetxt("mask.list", mask_list, fmt="%s")

    imagefiles = [
        outputDir + os.path.splitext(filenameInput)[0] + "_ps1_mosaic",
        outputDir + os.path.splitext(filenameInput)[0] + "_ps1_mosaic_mask",
    ]

    # Get pixel scale from input image header
    header = fits.getheader(inputimage)

    try:
        pixScale = abs(header["CDELT1"])
    except Exception:
        try:
            pixScale = abs(header["CD1_1"])
        except Exception:
            print(
                "Pixel scale could not be found in fits header.\n"
                "Expected keyword: CDELT1 or CD1_1"
            )
    pixScale = pixScale * 3600
    # print (inputimage, pixScale)
    crval1 = header["CRVAL1"]
    crval2 = header["CRVAL2"]
    # print (crval1, crval2)
    # print (header['CRPIX1'], header['CRPIX2'])
    imagesize = [header["NAXIS1"], header["NAXIS2"]]

    #  Force reference pixel to be in the center

    # File name to store the common header that will be shared by all
    # images in filelist
    point = "registration"
    # Delete if already exists
    rm_p(point + ".head")
    # First run swarp to create a .head file containing the shared header
    subprocess.call(
        [
            "swarp",
            "-HEADER_ONLY", "Y",
            "-IMAGEOUT_NAME", point + ".head",
            "-VERBOSE_TYPE", verbose,
        ]
        + [inputimage]
    )
    # Some keywords manipulation using sed
    subprocess.call(
        [
            "sed",
            "-i",
            "s/MJD-OBS/COMMENT/; s/EXPTIME/COMMENT/; s/GAIN   /COMMENT/; s/SATURATE/COMMENT /",
            point + ".head",
        ]
    )

    imalists = [["@" + "mosaic.list"], ["@" + "mask.list"]]
    for i, imagefile in enumerate(imagefiles):
        #  Remove mosaic if already exists
        rm_p(imagefile + ".fits")
        if "mask" in imagefile:
            subBackground = "N"
        else:
            subBackground = "Y"

        # Copy the common header in the .head file
        # So that it is read by sawrp for each image
        shutil.copy(point + ".head", imagefile + ".head")

        if useweight:
            subprocess.call(
                [
                    "swarp",
                    "-IMAGEOUT_NAME", imagefile + ".fits",
                    "-WEIGHTOUT_NAME", imagefile + ".weight.fits",
                    "-VERBOSE_TYPE", verbose,
                ]
                + imalists[i]
            )
        else:
            subprocess.call(
                [
                    "swarp",
                    "-IMAGEOUT_NAME", imagefile + ".fits",
                    "-SUBTRACT_BACK", subBackground,
                    "-COMBINE", "Y",
                    "-BACK_SIZE", "128",
                    "-BACK_FILTERSIZE", "3",
                    # '-CENTER_TYPE', 'MANUAL',
                    # '-CENTER', '%s, %s' % (crval1,crval2),
                    "-RESAMPLE", "Y",
                    #"-RESAMPLING_TYPE", "LANCZOS3",
                     '-RESAMPLING_TYPE', 'BILINEAR',
                    "-PIXELSCALE_TYPE", "MANUAL",
                    "-PIXEL_SCALE", str(pixScale),
                    # '-IMAGE_SIZE', '%s, %s' % (imagesize[0], imagesize[1]),
                    "-OVERSAMPLING", "0",
                    "-COMBINE_TYPE", "MEDIAN",
                    "-COPY_KEYWORDS", " PIXEL_SCALE",
                    "-VERBOSE_TYPE", verbose,
                ]
                + imalists[i]
            )
        rm_p(imagefile + ".head")
    #  Perform astrometric calibration of the mosaic with scamp
    scamp(
        imagefiles[0] + ".fits",
        config,
        useweight=False,
        CheckPlot=False,
        verbose=verbose,
    )
    # replace pixels == 0 with NaNs. Mostly the border, saturated pixels
    hdulist = fits.open(imagefiles[0] + ".fits")
    hdulist[0].data[hdulist[0].data == 0] = np.nan
    """
    # Add header
    hdulist[0].header['FILTER'] = band
    hdulist[0].header['PHOT_C'] = 25
    hdulist[0].header['PHOT_K'] = 0
    hdulist[0].header['PHOTFLAG'] = 'T'
    hdulist[0].header['EXPTIME'] = 1
    """
    hdulist[0].header["GAIN"] = 1
    hdulist[0].header["EXPTIME"] = 1
    hdulist[0].header.remove("SATURATE")

    hdulist.writeto(imagefiles[0] + ".fits", overwrite=True)

    #  Create a mask to propagate the nan pixels
    hdulist = fits.open(imagefiles[1] + ".fits")
    hdulist[0].data[hdulist[0].data > 0] = 1
    hdulist[0].data[np.isnan(hdulist[0].data)] = 1
    hdulist.writeto(imagefiles[1] + ".fits", overwrite=True)

    # for ima in file_list:
    #    rm_p(ima)
    # for ima in mask_list:
    #    rm_p(ima)
    rm_p("mosaic.list")
    rm_p("mask.list")
    rm_p("swarp.xml")
    rm_p(point + ".head")
    # rm_p('coadd.weight.fits')

    #  Add extension to files
    imagefiles = [i + ".fits" for i in imagefiles]

    return imagefiles



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu May  6 00:30:32 2021

@author: duverne, IJClab, Orsay, duverne@protonmail.com

Author: David Corre, Orsay, France, corre@iap.fr

"""

import argparse
import warnings
import os

from muphoten.utils import (
    load_config,
    list_files,
    make_results_dir,
    getpath,
    getTel)    
from muphoten.psfex import psfex

warnings.simplefilter(action="ignore", category=FutureWarning)


def main():

    path_muphoten = getpath()
    telescope_list = getTel()

    parser = argparse.ArgumentParser(
        usage="usage: %(prog)s data [data2 ... dataN] [options]",
        description="Compute PSF of astronomical images."
    )

    parser.add_argument(
        "--images",
        dest="path_image",
        required=False,
        type=str,
        help="Base path to repertory containing the images to analyse. "
    )
    
    # parser.add_argument(
    #     "--results",
    #     dest="path_results",
    #     required=False,
    #     type=str,
    #     default='muphoten_results',
    #     help="Base path to store the results. "
    #          "(Default: muphoten_results)"
    # )

    parser.add_argument(
        "--keep-old",
        "--keep",
        dest="keep",
        required=False,
        action="store_true",
        help="Keep previous results"
    )

    parser.add_argument(
        "--skip-processed",
        "--skip",
        dest="skip",
        required=False,
        action="store_true",
        help="Skip already processed files"
    )

    parser.add_argument(
        "--telescope",
        dest="telescope",
        choices=telescope_list,
        required=True,
        type=str,
        help="Alias for the available telescopes.",
    )

    parser.add_argument("--conv-filter",
                        dest="convFilter",
                        required=False,
                        default="default",
                        type=str,
                        help="Corresponds to FILTER_NAME keyword for sextractor"
                             "(without .conv)."
                             "\nDifferent filter available listed here: %s"
                        % path_muphoten + "/config/conv_kernels/"
                             "\n(Default: default)",)

    parser.add_argument(
        "--use-weight",
        dest="useweight",
        action="store_true",
        help="If set, use weight map. "
             "Must be same name as image with .weight.fits extension. "
             "(Default: False)",
    )

    parser.add_argument(
        "--verbose",
        dest="verbose",
        required=False,
        default="NORMAL",
        choices=["QUIET", "NORMAL", "FULL", "LOG"],
        type=str,
        help="Level of verbose, according to astromatic software. "
             "(Default: NORMAL)",
    )

    args = parser.parse_args()
    print(args.telescope)
    # Load config files for a given telescope
    config = load_config(args.telescope, args.convFilter)

    filenames = list_files(args.path_image)[0]
    print('Processing {} images'.format(len(filenames)))

    for image in filenames:

        filename = make_results_dir(image,
                                    directory='psf',
                                    keep=args.keep,
                                    skip=args.skip)
        print(filename)
        if not filename:
            print("%s is already processed, skipping. \n" % image)
            continue

            if not os.path.exists(filename):
                print("Pre-processing failed")
                continue

        psfex(filename, config, useweight=args.useweight,
              verbose=args.verbose, outLevel=2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import glob

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['astropy>=4.3.1',
                'astroquery>=0.4.4.dev7007',
                'Bottleneck>=1.3.2',
                'hjson>=3.0.2',
                'matplotlib>=3.4.3',
                'numpy>=1.21.2',
                'photutils==1.2.0',
                'regions>=0.5',
                'xmltodict>=0.12.0',
                'scikit-image>=0.18.3',
                'scipy>=1.7.1',
                'shapely']

setup_requirements = [ ]

test_requirements = [ ]

data_files = [("ps1_survey", ["muphoten/ps1_survey/ps1grid.fits"]),
              ("config", glob.glob("muphoten/config/*/*"))]

setup(
    author="Pierre-Alexandre Duverne, David Corre",
    author_email='duverne@protonmail.com',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="A photometry tool for transient astromy"
                "design for network of telescopes.",
    entry_points={'console_scripts': ['mu_bkg_subtraction=muphoten.cli.background_subtraction:main',
                                      'mu_mag_lim=muphoten.cli.mag_lim:main',
                                      'mu_photometry=muphoten.cli.photometry:main',
                                      'mu_psf=muphoten.cli.psf:main',
                                      'mu_sanitise=muphoten.cli.sanitise:main',
                                      'mu_subtraction=muphoten.cli.subtraction:main']},
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords=['photometry', 'transient', 'astronomy'],
    name='Muphoten',
    packages=find_packages(include=['muphoten', 'muphoten.*']),
    data_files=data_files,
    setup_requires=setup_requirements,
    # test_suite='tests',
    # tests_require=test_requirements,
    url='https://gitlab.in2p3.fr/icare/MUPHOTEN',
    version='0.1.0',
    zip_safe=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import subprocess
import shutil
import numpy as np
from astropy.io import fits
from astropy import wcs
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.table import Table
from copy import deepcopy
from muphoten.utils import rm_p, mkdir_p
from muphoten.astrometry import scamp


def registration(filelist, config, resultDir="", reference=None,
                 useweight=False, gain=1, normalise_exp=True,
                 verbose="NORMAL"):
    """Register images"""
    # Initialise lists used for creating output astropy table
    inim_list = []
    refim_list = []
    mask_list = []
    XY_lim = []
    in_lo = []
    in_up = []
    ref_lo = []
    ref_up = []
    gain_in = []
    gain_ref = []

    for i in range(len(filelist)):
        inim = filelist[i][0]
        refim = filelist[i][1]
        refim_mask = filelist[i][2]

        # Create list of images to register
        files = [inim, refim]
        if refim_mask is not None:
            files = files + [refim_mask]
        # Save them in a file to give it as an argument to swarp
        np.savetxt("register.list", files, fmt="%s")

        # Get pixel scale from input image header
        header = fits.getheader(inim)
        if "CDELT1" in header:
            pixScale = abs(float(header["CDELT1"])) * 3600
        elif ("CD1_1" and "CD2_1") in header:
            pixScale = np.sqrt(float(header["CD1_1"])**2 +
                               float(header["CD2_1"]**2)) * 3600
        else:
            raise ValueError('No information about pixel scale found.')

        imalists = ["@" + "register.list"]
        # File name to store the common header that will be shared by all
        # images in filelist
        point = "registration"
        # Delete if already exists
        rm_p(point + ".head")
        # First run swarp to create a .head file containing the shared header
        subprocess.call(
            [
                "swarp",
                "-HEADER_ONLY", "Y",
                "-IMAGEOUT_NAME", point + ".head",
                "-GAIN_DEFAULT", str(gain),
                # '-VERBOSE_TYPE', verbose] + imalists)
                "-VERBOSE_TYPE", verbose,
            ]
            # + [inim]
            + [refim]
        )
        # Some keywords manipulation using sed
        subprocess.call(
            [
                "sed",
                "-i",
                "s/MJD-OBS/COMMENT/; s/EXPTIME/COMMENT/; s/GAIN   /COMMENT/; s/SATURATE/COMMENT /",
                point + ".head",
            ]
        )
        outFiles = []
        # Run swarp to perform the registration on each image in filelist
        for j, ima in enumerate(files):

            path, filename_ext = os.path.split(ima)
            epoch = resultDir + os.path.splitext(filename_ext)[0] + \
                "_reg_%s" % i
            outFiles.append(epoch + ".fits")

            if "mask" in ima:
                subBackground = "N"
            else:
                subBackground = "Y"

            # use weight for PS1 image
            if reference == 'ps1' and 'rings_v3_skycell' in ima and \
                    'mask' not in ima:
                # weight_type = "MAP_WEIGHT"
                # weight_type = "MAP_VARIANCE"
                weight_type = "MAP_RMS"
                # weight_type = "NONE"

                weight_name = path + '/' + \
                    os.path.splitext(filename_ext)[0] + \
                    ".weight.fits"
            else:
                weight_type = "NONE"
                weight_name = path + '/' + \
                    os.path.splitext(filename_ext)[0] + \
                    ".weight.fits"

            # Copy the common header in the .head file
            # So that it is read by sawrp for each image
            shutil.copy(point + ".head", epoch + ".head")

            # Use bilinear to avoid artefact, but worst for
            # noise so would need to check in more details.
            subprocess.call(
                [
                    "swarp",
                    "-IMAGEOUT_NAME", epoch + ".fits",
                    "-WEIGHT_TYPE", weight_type,
                    "-WEIGHT_IMAGE", weight_name,
                    "-WEIGHTOUT_NAME", '.weight.fits',
                    # Arbitrary threshold.
                    # Pixels at the edsge after resampling are 0 so
                    # it is enough here
                    "-WEIGHT_THRESH", "0.1",
                    "-RESCALE_WEIGHTS", "N",
                    # '-GAIN_DEFAULT', str(gain),
                    "-FSCALE_KEYWORD", "NONE",
                    "-FSCALE_DEFAULT", "1, 1",
                    "-SUBTRACT_BACK", subBackground,
                    "-COMBINE", "Y",
                    "-COMBINE_TYPE", "MEDIAN",
                    "-BACK_SIZE", "128",
                    "-BACK_FILTERSIZE", "3",
                    "-RESAMPLE", "Y",
                    "-RESAMPLE_DIR", resultDir,
                    "-RESAMPLE_SUFFIX", '_test.fits',
                    "-PIXELSCALE_TYPE", "MANUAL",
                    "-PIXEL_SCALE", str(pixScale),
                    # '-CENTER', '%s, %s' % (header['CRVAL1'],
                    #                         header['CRVAL2']),
                    '-RESAMPLING_TYPE', 'LANCZOS3',
                    # "-RESAMPLING_TYPE", "BILINEAR",
                    # '-RESAMPLING_TYPE', 'NEAREST',
                    "-OVERSAMPLING", "0",
                    "-VERBOSE_TYPE", verbose,
                    "-COPY_KEYWORDS", "FILTER, DATE-OBS",
                ]
                + [ima]
            )

            # replace borders with NaNs in ref image if there are
            # any that are == 0,
            # hdulist=fits.open(epoch + '.fits')
            # hdulist[0].data[hdulist[0].data==0]=np.nan
            # hdulist.writeto(epoch + '.fits',overwrite=True)

            rm_p(epoch + ".head")
        rm_p(point + ".head")
        rm_p("register.list")
        rm_p("coadd.weight.fits")
        rm_p('swarp.xml')
        inim_regist = outFiles[0]
        refim_regist = outFiles[1]
        if refim_mask is not None:
            maskim_regist = outFiles[2]
        # else:
        #     maskim_regist = None
        else:
            maskref_regist = flag_bad_pixels(refim_regist).astype(int)
            header = fits.getheader(refim_regist)
            fits.writeto(os.path.join(os.path.dirname(inim_regist),
                                      'maskref_regist.fits'),
                         maskref_regist,
                         header, overwrite=True)
            maskim_regist = os.path.join(os.path.dirname(inim_regist),
                                        'maskref_regist.fits')

        print('Rescale and homogeneise mask maps for bad pixels.')
        # Rescale flux to 1s
        # In the future can try to rescale flux of ref image
        # to match input image.
        if normalise_exp:
            rescale_flux(inim_regist)
            rescale_flux(refim_regist)

        # Set masked pixels to same value
        mask_pix = flag_bad_pixels(inim_regist,
                                   mask_ref=maskim_regist,
                                   value=1e-30)
            
        # Update mask map
        _ = flag_bad_pixels(maskim_regist, value=1e8, mask_map=mask_pix)
        # Apply mask on ref data
        _ = flag_bad_pixels(refim_regist, mask_ref=maskim_regist, value=1e-30)

        print('Remove bad pixels on the edge.')
        # Take only part of image with data
        # This will decrease image size and speed up the substraction
        # Might also avoid probelm with masked values, depending on how
        # good they are deal with in hotpants.
        limits = keep_useful_area(inim_regist, image_ref=refim_regist)
        _ = keep_useful_area(maskim_regist, limits_force=limits)

        # Perform a second time, as the edge are not straight, we can still
        # remove some pixels after the first cut.
        limits = keep_useful_area(inim_regist, image_ref=refim_regist)
        _ = keep_useful_area(maskim_regist, limits_force=limits)

        # Get info to tune hotpants parameters
        filelist_regist = [inim_regist, refim_regist, maskim_regist]
        hotpants_info = get_hotpants_info(filelist_regist, config, verbose)

        inim_list.append(outFiles[0])
        refim_list.append(outFiles[1])
        mask_list.append(maskim_regist)
        XY_lim.append(hotpants_info[0])
        in_lo.append(hotpants_info[1][0])
        in_up.append(hotpants_info[1][1])
        ref_lo.append(hotpants_info[1][2])
        ref_up.append(hotpants_info[1][3])
        gain_in.append(hotpants_info[2][0])
        gain_ref.append(hotpants_info[2][1])

    info = Table(
        [
            inim_list,
            refim_list,
            mask_list,
            XY_lim,
            in_lo,
            in_up,
            ref_lo,
            ref_up,
            gain_in,
            gain_ref,
        ],
        names=[
            "inim",
            "refim",
            "mask",
            "XY_lim",
            "in_lo",
            "in_up",
            "ref_lo",
            "ref_up",
            "gain_in",
            "gain_ref",
        ],
    )
    return info


def flag_bad_pixels(image, mask_ref=None, value=1e-30, mask_map=None):
    """
    Set masked pixels to same value, to homogeneise.
    1e-30 is the value used by hotpants to ignore pixels.
    """
    hdulist1 = fits.open(image)

    if mask_map is not None:
        # Apply an additional mask map to data
        hdulist1[0].data[mask_map] = value

    if mask_ref is not None:
        hdulist2 = fits.open(mask_ref)
        mask = hdulist2[0].data != 0.0
        hdulist2.close()

    if value == 1e-30:
        # Used to flag bad pixels in science and ref images by hotpants
        # Can actually be around the 1e-30 if resampled have been
        # done iwith weight on science image.
        hdulist1[0].data[hdulist1[0].data == 0] = value
        # Flag bad pixels from the mask
        if mask_ref is not None:
            hdulist1[0].data[mask] = value
    else:
        # For the mask map, pixels with high values are ignored.
        # 0 means they are not masked, i.e. good.
        hdulist1[0].data[hdulist1[0].data != 0] = value

    output_mask = hdulist1[0].data == value
    hdulist1.writeto(image, overwrite=True)

    return output_mask


def rescale_flux(image):
    """
    Rescale flux scale to 1s.
    In the future can also compute the flux factor scaling
    to match science and input image flux scale.
    """

    # Normalise pixel flux to 1s.
    hdulist = fits.open(image)
    hdulist[0].data = hdulist[0].data / hdulist[0].header["EXPTIME"]
    # Try to update headers that are affected by this change.
    # RN and DC standing for Read Noise and Dark Current are currently
    # not kept during sanitising of the header. Need to update it.
    keywords = ['SATURATE', 'RN', 'DC']
    for key in keywords:
        try:
            hdulist[0].header[key] = (
                hdulist[0].header[key] /
                hdulist[0].header["EXPTIME"]
            )
        except BaseException:
            pass
    # Set Exposure time to 1s from now on.
    hdulist[0].header["EXPTIME"] = 1
    hdulist.writeto(image, overwrite=True)


def keep_useful_area(image, image_ref=None, limits_force=None):
    """Keep only part of image with non masked pixels"""
    hdulist1 = fits.open(image)
    if image_ref is not None:
        hdulist2 = fits.open(image_ref)

    if limits_force is None:
        ymin_im, xmin_im = np.min(np.where(hdulist1[0].data > 1e-30), axis=1)
        ymax_im, xmax_im = np.max(np.where(hdulist1[0].data > 1e-30), axis=1)

        if image_ref is None:
            # Do it on a single image
            xmin = xmin_im
            xmax = xmax_im
            ymin = ymin_im
            ymax = ymax_im

        else:
            ymin_ref, xmin_ref = np.min(
                np.where(hdulist2[0].data > 1e-30), axis=1)
            ymax_ref, xmax_ref = np.max(
                np.where(hdulist2[0].data > 1e-30), axis=1)

            xmin = np.max([xmin_ref, xmin_im])
            xmax = np.min([xmax_ref, xmax_im])
            ymin = np.max([ymin_ref, ymin_im])
            ymax = np.min([ymax_ref, ymax_im])

    else:
        xmin, xmax, ymin, ymax = limits_force
    hdulist1[0].data = hdulist1[0].data[ymin:ymax, xmin:xmax]

    # Need to update center position info in header
    # Not CRVAL ?? Need to check
    xcenter1, ycenter1 = (hdulist1[0].header["CRPIX1"],
                          hdulist1[0].header["CRPIX2"])

    xcenter, ycenter = int(xcenter1 - xmin) + 1, int(ycenter1 - ymin) + 1
    hdulist1[0].header["CRPIX1"] = xcenter
    hdulist1[0].header["CRPIX2"] = ycenter

    hdulist1.writeto(image, overwrite=True)

    if image_ref is not None:
        hdulist2[0].data = hdulist2[0].data[ymin:ymax, xmin:xmax]

        # Need to update center position info in header
        xcenter1, ycenter1 = (hdulist2[0].header["CRPIX1"],
                              hdulist2[0].header["CRPIX2"])

        xcenter, ycenter = int(xcenter1 - xmin) + 1, int(ycenter1 - ymin) + 1
        hdulist2[0].header["CRPIX1"] = xcenter
        hdulist2[0].header["CRPIX2"] = ycenter

        hdulist2.writeto(image_ref, overwrite=True)

    return [xmin, xmax, ymin, ymax]


def get_hotpants_info(filelist, config, verbose):
    """Get some information for tuning hotpants
    """
    # Get the min and max pix with non 0 values
    # to delimate the new frame
    imdata, imheader = fits.getdata(filelist[0], header=True)
    refdata, refheader = fits.getdata(filelist[1], header=True)

    inmin = -10  # np.nanmin(imdata)
    if "SATURATE" in imheader:
        inmax = 0.9 * imheader["SATURATE"]
    else:
        inmax = 0.9 * np.nanmax(imdata)

    refmin = -10  # np.nanmin(refdata)
    if "SATURATE" in refheader:
        refmax = 0.9 * refheader["SATURATE"]
    else:
        refmax = 0.9 * np.nanmax(refdata)

    imgain = imheader.get('GAIN', 1)
    refgain = refheader.get('GAIN', 1)

    if imgain == 0 or imgain > 10:
        imgain = 1.0

    if refgain == 0 or refgain > 10:
        refgain = 1.0

    # Not needed anymore since using the weight output of swarp.
    # That flags correctly the pixels at the edge after the resampling.
    xmin = None
    ymin = None
    xmax = None
    ymax = None

    return [[xmin, xmax, ymin, ymax], [
        inmin, inmax, refmin, refmax], [imgain, refgain]]


if __name__ == "__main__":

    im1 = "Test_Atlascow/muphoten_results/ATLAS18qqn-S001-R001-C001-SDSS_g.fits"
    im2 = "ps1_rescaled/ps1_mosaic.fits"
    registration(im1, im2, False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 21 22:32:34 2021

@author: duverne, IJClab, Orsay, duverne@protonmail.com
"""

import sys
import argparse
from muphoten.sanitise import add_filter, add_observer, sanitise_fits
from astropy.io import fits
from muphoten.utils import list_files

"""Muphoten script for preparing and making the header right."""

def main():
    
    parser = argparse.ArgumentParser(description="Performing photometry of\
                                     transient objects.")
    
    parser.add_argument("--images",
                    dest="images",
                    required=True,
                    type=str,
                    help="Path to images.")

    parser.add_argument("--telescope",
                    required=True,
                    type=str,
                    help="Telescope that acquired the images.")
    
    parser.add_argument("--filter",
                    required=True,
                    type=str,
                    help="Filter used to acquire the images.")
    args = parser.parse_args()

    images = list_files(args.images, get_subdirs=False)

    for i, image in enumerate(images):
        hdulist = fits.open(image)
        header = hdulist[0].header

        if 'TELESCOPE' not in header.items():
            add_observer(image, args.telescope)
        if 'FILTER' not in header.items():
            add_filter(image, args.filter)
        try:
            sanitise_fits(image)
        except Exception:
            hdulist.verify("fix")
            sanitise_fits(image)

    return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Tue Feb  2 16:37:19 2021

@author: duverne
"""

import numpy as np
from astropy import units as u, coordinates as coord
from astroquery.vizier import Vizier
from astropy.coordinates import Angle
import matplotlib.pyplot as plt
import muphoten.catalog as cata
from muphoten.background import sub_background
from muphoten.photometry import source_detection
from photutils import source_properties

def get_sky_coord_center(data, wcs):
    """
    Get the center of the image in WCS.

    """

    x_center, y_center = int(np.shape(data)[0]/2), int(np.shape(data)[1]/2)
    coord_center = wcs.all_pix2world(y_center, x_center, 0)
    center = coord.SkyCoord(coord_center[0], coord_center[1],
                            unit=(u.deg, u.deg), frame='icrs')
    return center

def get_fov(data, wcs):
    """
    Get the FoV nformation about the image

    """
    
    size = np.shape(data)
    origin = wcs.all_pix2world(0, 0, 0)
    origin = coord.SkyCoord(origin[0], origin[1],
                            unit=(u.deg, u.deg), frame='icrs')
    
    sup = wcs.all_pix2world(0, size[0], 0)
    sup = coord.SkyCoord(sup[0], sup[1],
                         unit=(u.deg, u.deg), frame='icrs')
    
    inf = wcs.all_pix2world(size[1], 0, 0)
    inf = coord.SkyCoord(inf[0], inf[1],
                         unit=(u.deg, u.deg), frame='icrs')
    
    
    height = origin.separation(sup).deg * u.deg
    width = origin.separation(inf).deg * u.deg
    
    return width, height

def get_region(data, wcs, band='gmag'):
    """
    Get all the objects found in the FoV of the image by Pan-STARRS.

    Parameters
    ----------
    data : numpy array
        The image array.
    wcs : Astropy WCS object
        Astrometry fo the image.
    band : string, optional
        Filter used to acquire the image. The default is 'gmag'.

    Returns
    -------
    region : Astropy array
        The table withthe information about the objects in the FoV.

    """
    
    johnson = ['BMag', 'VMag', 'RMag', 'IMag']
    center = get_sky_coord_center(data, wcs)
    width, height = get_fov(data, wcs)
    Vizier.ROW_LIMIT = 999999999
    Vizier.TIMEOUT=1000.
    region = Vizier.query_region(center, width = Angle(width, "deg"), 
                                 height = Angle(height, "deg"),
                                 catalog = 'II/349/ps1')
    region = region[0]
    region['xcentroid'], region['ycentroid'] = wcs.all_world2pix(region['RAJ2000'],
                                                                 region['DEJ2000'],
                                                                 0)
    if band in johnson:
        region = cata.from_PS2Johnson(band, region)

    if band=='cmag':
        region = cata.from_PS2clear(band, region)

    return region

def detect_objects(data, wcs, band, mu_conf):   
    """
    Find all the objects detected in the image, change the photometric
    system if necessary.

    Parameters
    ----------
    data : numpy array
        image data.
    header : Astropy header
        Image header.
    band : string
        Filter used to acquire the image.
    mu_conf : dictonnary
        Muphoten configuration for the analysis.
    path : path, string
        Path where to save the background image.
    im_name : string
        image name.

    Returns
    -------
    detected_objects : Astropy Table
        Tabme with information about the objects in the FoV.

    """
    
    johnson = ['BMag', 'VMag', 'RMag', 'IMag']
    data_clean, background, rms = sub_background(data, None, None,
                                                 n_sigma=mu_conf['bkg_n_sigma'],
                                                 box_size=mu_conf['bkg_box'],
                                                 bkg_estimator=mu_conf['bkg_estimator'],
                                                 filter_size=mu_conf['bkg_filter_size'])
    #Detection of the sources
    segmentation = source_detection(data, background, rms,
                                    n_sigma=mu_conf['det_n_sigma'],
                                    deblend=mu_conf['det_deblend'],
                                    threshold_level=mu_conf['det_treshold'],
                                    npixels=mu_conf['det_npixels'],
                                    contrast=mu_conf['contrast'])
    cat = source_properties(data_clean, segmentation,
                            background=background,
                            wcs=wcs)

    columns = ['id', 'xcentroid', 'ycentroid', 'sky_centroid']
    table = cat.to_table(columns=columns)
    table['idx'] = np.arange(len(table))
    coord_sources = cata.source_coordinates(table)
    detected_objects = cata.xmatch(coord_sources, 'II/349/ps1', 5.0)
    if band in johnson:
        detected_objects = cata.from_PS2Johnson(band, detected_objects)

    if band=='cmag':
        detected_objects = cata.from_PS2clear(band, detected_objects)

    return detected_objects

def build_ratio(region, detected_objects, band,
                precision=0.2, interval=[15., 22.]):
    """
    Building if the ratio of the detected sources in the image and the objects
    in the catalog.

    Parameters
    ----------
    region : Astropy table
        Table of the sources detected in the FoV by Pan-STARRS.
    detected_objects : Astropy Table
        Table of the sources detected in the image.
    band : string
        Filter used to acquire the image.
    precision : float, optional
        Bin size. The default is 0.2.
    interval : list, optional
        Interval of magnitude where to find the limit magnitude.
        The default is [15., 22.].

    Returns
    -------
    histo_obj : numpy histogram
        Histogram of the sources detected in the image.
    histo_region : numpy histogram
        Histogram of the sources detected in Pan-STARRS.
    mag_lim : numpy histogram
        Limit magnitude of the image.

    """
    
    magmax,  magmin = np.max(region[band]), np.min(region[band])
    rg=(magmin, magmax)
    nbin = int((magmax - magmin)/precision)
    
    mask = (region[band] > interval[0]) & (region[band] < interval[1]) 
    region = region[mask]
    mask = (detected_objects[band] < np.max(region[band])) \
            & (detected_objects[band] > interval[0])
    detected_objects = detected_objects[mask]

    histo_region = np.histogram(region[band], bins= nbin, range=rg)
    histo_obj = np.histogram(detected_objects[band],
                             bins= nbin, range=rg)
    
    mag_lim = np.zeros(np.size(histo_region[0]), dtype=np.float64)
    
    for i in range(len(histo_region[0])):
        if histo_region[0][i] != 0:
            ratio = histo_obj[0][i] / histo_region[0][i]
            if ratio >= 1.0 :
                mag_lim[i] = 1.0
            else:
                mag_lim[i] = ratio

    return histo_obj, histo_region, mag_lim


def do_plot_histo(region, detected_objects, mag_lim, lim, band,
                  precision=0.2, threshold=0.5, name='mag_lim'):
    """
    Plot the histrograms of the sources detcted in the image and by Pan-STARRS
    in the FoV.

    Parameters
    ----------
    region : Astropy table
        Table of the sources detected in the FoV by Pan-STARRS.
    detected_objects : Astropy Table
        Table of the sources detected in the image.
    mag_lim : numpy histogram
        Limit magnitude of the image.
    lim : float
        Value of the limit magnitude of the image.
    band : string
        Filter used to acquire the image.
    precision : float, optional
        Bin size. The default is 0.2.
    threshold : float, optional
        Threshold of the ratio to define where the limit magnitude is reach.
        The default is 0.5.
    name : string, optional
        name for the plot save. The default is 'mag_lim'.

    Returns
    -------
    None.

    """
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    label = 'Limit Magnitude = {:.1f} $\pm$ {} mag'.format(lim, precision)
    magmax,  magmin = np.max(region[band]), np.min(region[band])
    rg=(magmin, magmax)
    nbin = int((magmax - magmin)/precision)
    ax.axvline(lim, label = label,
               color='grey', linestyle='dashed', linewidth=3)
    ax.hist(detected_objects[band], bins=nbin, alpha=0.5,
            range=rg, label=r'image')
    ax.hist(region[band], bins= nbin, alpha=0.5,
            range=rg, label=r'Pan-STARRS')
    ax.legend()
    ax.set_title(r'Limit Magnitude')
    ax.set_xlabel(r'Magnitude [mag]')
    ax.set_ylabel(r'Number of detected objects')
    # label = r'Limit Magnitude = ' + '%.2f' %lim
    ax.axvline(lim, label = label,color='grey', linestyle='dashed', linewidth=3)
    
    if name is not None:
        fig.savefig(name + '.png')
        plt.close(fig)
    
def do_plot_ratio(mag_lim, lim, region, index, threshold=0.5, precision = 0.2,
                  name='ratio', title=r'Limit Magnitude Ratio'):
    """
    Plot the ratio histogram.

    Parameters
    ----------
    mag_lim : numpy histogram
        Limit magnitude of the image.
    lim : float
        Value of the limit magnitude.
    region : Astropy table
        Table of the sources detected in the FoV by Pan-STARRS.
    index : int
        Index of the histogram where the limit magnitude is reached.
    threshold : float, optional
        Threshold of the ratio to define where the limit magnitude is reach.
        The default is 0.5.
    name : string, optional
        name for the plot save. The default is 'ratio'.
    title : string, optional
        Title of the plot. The default is r'$Limit$ $Magnitude$ $Ratio$'.

    Returns
    -------
    None.

    """
    mask = (mag_lim[:index+1]==0.)
    y = np.ma.array(mag_lim[:index+1], mask=mask).compressed()
    x = np.ma.array(region[1][:index+1], mask=mask).compressed()
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    label = 'Limit Magnitude = {:.1f} $\pm$ {} mag'.format(lim, precision)
    ax.axvline(lim, label = label, color='grey',
               linestyle='dashed', linewidth=3)
    ax.plot(x, y, linestyle='none', marker = 'o', color='blue', ms=8)
    ax.plot(region[1][index+1:-1], mag_lim[index+1:], 
            linestyle='none', marker = 'o', color='blue', ms=8)
    ax.axhline(threshold, label = r'threshold', color='black',
               linestyle='-', linewidth=3)
    ax.set_title(title)
    ax.title.set_size(20)
    ax.set_xlabel(r'Magnitude [mag]')
    ax.set_ylabel(r'ratio of detected objects')
    ax.xaxis.label.set_size(15)
    ax.yaxis.label.set_size(15)
    ax.grid(b=True, which='major', color='#666666', linestyle='-')
    ax.minorticks_on()
    ax.grid(b=True, which='minor', color='#999999', linestyle='-', alpha=0.5)
    ax.legend(prop={"size":15})
    if name is not None:
        fig.savefig(name + '.png')
        plt.close(fig)

def get_mag_lim(mag_lim, detected_objects, precision=0.2, threshold=0.5):
    """
    Get the limit magnitude of the image.

    Parameters
    ----------
    mag_lim : numpy histogram
        Limit magnitude of the image.
    detected_objects : Astropy Table
        Table of the sources detected in the image.
    precision : float, optional
        Bin size. The default is 0.2.
    threshold : float, optional
        Threshold of the ratio to define where the limit magnitude is reach.
        The default is 0.5.

    Returns
    -------
    lim : float
        Value of the limit Magnitude.
    idx : int
        Index of the histogram where the limit magnitude is reached.

    """
    index=0
    # threshold=0.5
    mask=mag_lim>0.
    for i in range(len(mag_lim[mask])):
        sup = mag_lim[mask][i:]>threshold
        if sup.any():
            index+= 1
        else:
            break
    value = mag_lim[mask][index-1]
    idx = np.max(np.where(mag_lim==value))
    lim = round(detected_objects[1][idx].astype('float16') + 0.5 * precision, 2)

    return lim, idx


def compute_mag_lim(data, wcs, band, mu_conf, precision, interval, threshold,
                    path_ratio, path_histo):
    """
    Main function to compute the limit magnitude and plot the ratio.

    Parameters
    ----------
    data : numpy array
        images array.
    wcs : astropy WCS
        Astrometry information.
    band : str
        Filter used to acquire the image.
    mu_conf : dict
        muphoten configuration.
    precision : float
        bin size.
    interval : list
        interval where to find the limit magnitude.
    threshold : float, optional
        Threshold of the ratio to define where the limit magnitude is reach.
    path_ratio : string
        Where to save the plot.

    Returns
    -------
    lim : TYPE
        DESCRIPTION.
    precision : TYPE
        DESCRIPTION.

    """
    width, height = get_fov(data, wcs)
    region = get_region(data, wcs, band)
    all_star_detected = detect_objects(data, wcs, band, mu_conf)
    detected_objects, hist_region, mag_lim = build_ratio(region,
                                                    all_star_detected,
                                                    band,
                                                    precision=precision,
                                                    interval=interval)
    lim, index = get_mag_lim(mag_lim, hist_region,
                             precision=precision,
                             threshold=threshold)
    do_plot_ratio(mag_lim,
                  lim,
                  hist_region,
                  index,
                  threshold,
                  precision,
                  name=path_ratio)

    do_plot_histo(region,
                  all_star_detected,
                  mag_lim,
                  lim,
                  band,
                  name=path_histo)

    return lim, precision
